/*****************************************************************************
 * va.c: Video Acceleration API for avcodec
 *****************************************************************************
 * Copyright (C) 2011, M2X BV
 * $Id: 21dd5050cdb909175ec1628cb3ed6eb2d99767a3 $
 *
 * Authors: Jean-Paul Saman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_fourcc.h>
#include <assert.h>

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#else
#   include <avcodec.h>
#endif

#include "avcodec.h"
#ifdef HAVE_AVCODEC_VAAPI
#  include <vlc_xlib.h>
#  include <libavcodec/vaapi.h>
#  include <va/va.h>
#  include <va/va_x11.h>
#endif
#include "va.h"
#include "vaapi.h"

#ifdef HAVE_AVCODEC_VAAPI

static vlc_mutex_t vlc_va_conn_lock = VLC_STATIC_MUTEX;

static void vlc_va_lock(void)
{
    vlc_mutex_lock(&vlc_va_conn_lock);
}

static void vlc_va_unlock(void)
{
    vlc_mutex_unlock(&vlc_va_conn_lock);
}

/* */
static vlc_va_surface_t *alloc_surfaces( vlc_va_conn_t *conn,
                                         const int width, const int height,
                                         const int fourcc, unsigned int *requested )
{
    assert( *requested > 0 );

    VASurfaceID *p_ids = (VASurfaceID *)calloc(*requested, sizeof(VASurfaceID));
    if( p_ids == NULL )
        return NULL;

    VAStatus status = vaCreateSurfaces( conn->p_display, width, height, fourcc,
                                        *requested, p_ids );
    if( status != VA_STATUS_SUCCESS )
    {
        free( p_ids );
        return NULL;
    }

    vlc_va_surface_t *p_surfaces = calloc(*requested, sizeof(vlc_va_surface_t));
    if( p_surfaces == NULL )
    {
        free( p_ids );
        return NULL;
    }

    for( unsigned int i = 0; i < *requested; i++ )
    {
        p_surfaces[i].i_id = p_ids[i];
        vlc_mutex_init(&p_surfaces[i].lock);
    }

    free( p_ids );
    return p_surfaces;
}

static vlc_va_surface_t *realloc_surfaces( vlc_va_conn_t *conn,
                                           const int width, const int height,
                                           const int fourcc, unsigned int *requested )
{
    assert( *requested > 0 );

    vlc_va_surface_t *p_surfaces = (vlc_va_surface_t *)realloc( conn->pool.p_surfaces,
                                                                (*requested) * sizeof(vlc_va_surface_t) );
    if( p_surfaces == NULL )
        return NULL;

    const unsigned int count = *requested - conn->pool.count;
    assert( count > 0 );

    VASurfaceID *ids = calloc(count, sizeof(VASurfaceID));
    if( ids == NULL )
        return NULL;
    VAStatus status = vaCreateSurfaces( conn->p_display, width, height, fourcc,
                                        count, ids );
    if( status != VA_STATUS_SUCCESS )
        abort(); /* FATAL ERROR */

    for( unsigned int i = 0; i < count; i++ )
    {
        vlc_va_surface_t *p = &p_surfaces[conn->pool.count + i];
        p->i_id = ids[i];
        vlc_mutex_init(&p->lock);
    }
    conn->pool.count += count;
    assert(conn->pool.count == *requested);
    free(ids);
    return p_surfaces;
}

static bool create_surfaces( vlc_va_conn_t *conn,
                                    const int width, const int height,
                                    const int fourcc, const unsigned int num )
{
    assert(conn);

    if( (width <= 0) || (height <= 0) || (num <= 0) )
        return false;

    if( conn->i_ref_count > 1 )
    {
        assert( conn->pool.p_surfaces != NULL );

        if( num > conn->pool.count )
        {
            assert(0);
            /* FIXME: probably wise to destroy decoding context first
               and then recreate it */
            unsigned requested = num;
            realloc_surfaces( conn, width, height, fourcc, &requested );
            assert( num == requested );
        }
        return true;
    }

    assert( conn->pool.p_surfaces == NULL );
    assert( conn->pool.count == 0 );

    unsigned requested = num;
    conn->pool.p_surfaces = alloc_surfaces( conn, width, height,
                                            fourcc, &requested );
    assert( num == requested );
    conn->pool.current = 0;
    if( conn->pool.p_surfaces )
    {
        conn->pool.count = requested;
        return true;
    }
    return false;
}

static void destroy_surfaces( vlc_va_conn_t *conn )
{
    assert( conn );
    assert( conn->pool.count > 0 );
    assert( conn->pool.p_surfaces );

    if( conn->i_ref_count > 1 )
        return;

    for( unsigned int i = 0; i < conn->pool.count; i++ )
    {
        vlc_va_surface_t *surface = &conn->pool.p_surfaces[i];

        vlc_mutex_lock(&surface->lock);
        if( surface->i_refcount > 1 )
            abort();

#if VA_CHECK_VERSION(0,31,0)
        if( vaSyncSurface( conn->p_display, surface->i_id ) )
#else
        if( vaSyncSurface( conn->p_display, conn->i_context_id, surface->i_id ) )
#endif
        {
            fprintf(stderr, "Fatal error reboot your system\n" );
            abort();
        }

        surface->i_refcount --;
        assert(surface->i_refcount == 0);

        vaDestroySurfaces( conn->p_display, &surface->i_id, 1 );
        vlc_mutex_unlock(&surface->lock);

        vlc_mutex_destroy(&surface->lock);
    }

    free(conn->pool.p_surfaces);

    conn->pool.count = 0;
    conn->pool.p_surfaces = NULL;
}

static vlc_va_surface_t *get_surface(vlc_va_conn_t *conn)
{
    assert( conn );
    assert( conn->pool.count > 0 );
    assert( conn->pool.p_surfaces );

    unsigned int next = conn->pool.current;
    unsigned int tries = 0;
    while (tries < conn->pool.count)
    {
        VAStatus status;
        VASurfaceStatus ready;
        vlc_va_surface_t *surface;

        surface = &conn->pool.p_surfaces[next];

        vlc_mutex_lock(&surface->lock);
        VASurfaceID id = surface->i_id;
        status = vaQuerySurfaceStatus(conn->p_display, id, &ready);
        if ((status == VA_STATUS_SUCCESS) &&
            (ready == VASurfaceReady) &&
            (surface->i_refcount == 0))
        {
            surface->i_refcount++;
            conn->pool.current = (next + 1) % conn->pool.count;
            vlc_mutex_unlock(&surface->lock);
            return surface;
        }
        vlc_mutex_unlock(&surface->lock);

        next = (next + 1) % conn->pool.count;
        tries++;
    }
    return NULL;
}

static void release_surface(vlc_va_conn_t *conn, vlc_va_surface_t *surface)
{
    assert( conn );
    assert( conn->pool.count > 0 );
    assert( conn->pool.p_surfaces );
    assert( surface );

    vlc_mutex_lock(&surface->lock);
    if (surface->i_refcount == 0)
        abort();

    VAStatus status;
    VASurfaceStatus ready;
    VASurfaceID id = surface->i_id;
    status = vaQuerySurfaceStatus(conn->p_display, id, &ready);
    if ((status == VA_STATUS_SUCCESS) &&
        (ready == VASurfaceReady))
    {
        surface->i_refcount--;
        assert(surface->i_refcount == 0);
    }
    vlc_mutex_unlock(&surface->lock);
}

static void release_surface_by_id(vlc_va_conn_t *conn, VASurfaceID id)
{
    assert( conn );
    assert( conn->pool.count > 0 );
    assert( conn->pool.p_surfaces );

    for( unsigned int i = 0; i < conn->pool.count; i++ )
    {
        vlc_va_surface_t *p_surface = &conn->pool.p_surfaces[i];
        if (!p_surface)
            abort();

        if( p_surface->i_id == id )
            release_surface(conn, p_surface);
    }
}

/* Global VAAPI connection state */
static vlc_va_conn_t vlc_va_conn = {
    .x11 = NULL,
    .p_display = 0,
    .b_direct_rendering = false,
    /* libva version */
    .i_version_major = 0,
    .i_version_minor = 0,
    /* ref counting */
    .i_ref_count = 0,
    /* locking functions */
    .lock = vlc_va_lock,
    .unlock = vlc_va_unlock,
    /* surfaces - NOTE: must be called with lock held */
    .pool = {
        .count = 0,
        .current = 0,
        .p_surfaces = NULL,
    },
    .i_context_id = VA_INVALID_ID,
    .create_surfaces = create_surfaces,
    .destroy_surfaces = destroy_surfaces,
    .get_surface = get_surface,
    .release_surface = release_surface,
    .release_surface_by_id = release_surface_by_id,
};

/* */
static vlc_va_conn_t *vlc_va_get_conn( void )
{
    return (vlc_va_conn_t *) &vlc_va_conn;
}

static bool uses_direct_mapping(vlc_va_conn_t *conn)
{
    VADisplayAttribute attr;
    VAStatus status;

#if VA_CHECK_VERSION(0,34,0)
    attr.type  = VADisplayAttribRenderMode;
    attr.flags = VA_DISPLAY_ATTRIB_GETTABLE;

    status = vaGetDisplayAttributes(conn->p_display, &attr, 1);
    if (status == VA_STATUS_SUCCESS)
        return !(attr.value & (VA_RENDER_MODE_LOCAL_OVERLAY|
                               VA_RENDER_MODE_EXTERNAL_OVERLAY));
#else
    /* Does the driver use DirectSurface mode then we cannot use the
       same surface while it is still being in the display pipeline. */
    attr.type  = VADisplayAttribDirectSurface;
    attr.flags = VA_DISPLAY_ATTRIB_GETTABLE;

    status = vaGetDisplayAttributes(conn->p_display, &attr, 1);
    if (status == VA_STATUS_SUCCESS)
        return !attr.value;
#endif
    return false;
}

vlc_va_conn_t *vlc_va_Initialize( const char *display )
{
    vlc_mutex_lock( &vlc_va_conn_lock );

    /* connect global to caller */
    vlc_va_conn_t *conn = vlc_va_get_conn();
    assert( conn );
    if( conn->i_ref_count > 0 )
    {
        conn->i_ref_count++;
        vlc_mutex_unlock( &vlc_va_conn_lock );
        return conn;
    }

    /* X11 display */
    conn->x11 = XOpenDisplay(display);
    if( !conn->x11 )
        goto error;

    /* Create a VA display */
    conn->p_display = vaGetDisplay(conn->x11);
    if( !conn->p_display )
        goto error;

    if( vaInitialize( conn->p_display, &conn->i_version_major, &conn->i_version_minor ) )
        goto error;

    conn->i_ref_count++;
    conn->b_direct_rendering = uses_direct_mapping(conn);
    vlc_mutex_unlock( &vlc_va_conn_lock );
    return conn;

error:
    if( conn->x11 )
        XCloseDisplay( conn->x11 );
    vlc_mutex_unlock( &vlc_va_conn_lock );
    return NULL;
}

void vlc_va_Terminate( vlc_va_conn_t *conn )
{
    vlc_mutex_lock( &vlc_va_conn_lock );
    assert( conn );
    assert( conn->i_ref_count > 0 );

    conn->i_ref_count--;
    if( conn->i_ref_count == 0 )
    {
        assert( conn->pool.count == 0 );
        assert( conn->pool.p_surfaces == NULL );

        assert( conn->p_display );
        vaTerminate( conn->p_display );
        XCloseDisplay( conn->x11 );

        /* Reset values */
        conn->x11 = NULL;
        conn->p_display = 0;
        conn->i_version_major = conn->i_version_minor = 0;
        conn->i_ref_count = 0;
        conn = NULL;
    }

    vlc_mutex_unlock( &vlc_va_conn_lock );
}

#endif
